#!/usr/bin/env python
from WSRTrecipe import *
from job_parser import *
import os, os.path, time, threading, types

NewJob    = 1
UpdateJob = 2

##--------------------- Server Code, should maybe be in separate module? --------------------------
class server_wrapper(threading.Thread):
    """Base class to create a server to listen for new jobs in a separate thread.
    Mainly implements how to talk with th pipeline manager."""
    def __init__(self, server): ## might need to be overridden
        threading.Thread.__init__(self)
        self.server    = server
        self.stop      = False
        self.stop_lock = threading.Lock()
        self.jobs      = [] ## list of received new/updated jobs. Main thread will remove them
        self.jobs_lock = threading.Lock()  ## lock to make sure only one is using this list
        self.event     = threading.Event() ## event to signal the main thread
    
    def run(self):
        """Empty function override in a subclass"""
        pass ## override in sub-class
        
    def new_job(self, fileName):
        """Send an event to the pipeline manager that a new job is available."""
        self.jobs_lock.acquire()
        self.jobs.append((NewJob, fileName))
        self.event.set() ## signal the main thread we've done something
        self.event.clear()
        self.jobs_lock.release()

    def update_job(self, exportID, status):
        """Send an event to the pipeline manager that a job needs to be updated."""
        self.jobs_lock.acquire()
        self.jobs.append((UpdateJob, exportID, status))
        self.event.set() ## signal the main thread we've done something
        self.event.clear()
        self.jobs_lock.release()

    def serve_forever(self):
        """Accept requests until the pipeline manager signals to stop."""
        self.server.socket.settimeout(60)
        while 1:
            #? self.stop_lock.acquire() doesn't seem necessary
            if self.stop:
                #? self.stop_lock.release() doesn't seem necessary
                break
            #? self.stop_lock.release() doesn't seem necessary
            self.server.handle_request()
##--------------------- End Server Code --------------------------

class pipeline_manager(WSRTrecipe):
    def __init__(self):
        WSRTrecipe.__init__(self)
        ## inputs
        self.inputs['ConfigurationFile'] = 'pipeline_manager_config' ## without the .py
        self.inputs['NewJobsDirectory']  = ''
        self.inputs['JobsDirectory']     = ''
        self.inputs['LogDirectory']      = ''
        # We have no outputs
        ## Help text
        self.helptext = """
        Script to run as a server for executing individial jobs
        ConfigurationFile should be given without the '.py' extention.
        NewJobs is where new jobs should be written, the JobsDirectory
        contains unfinished Jobs"""
        
        self.jobs            = []
        self.parser          = job_parser()
        self.parser.messages = self.messages
        self.idle            = 0
    
    ## Code to generate results ---------------------------------------------
    def startup(self):
        """Tell the user we stared, read the configuration and try to read unfinished jobs from JobDirectory"""
        print 'WSRT pipeline manager version 0.4'
        print 'Press Ctrl-C to abort'
        exec(eval("'from %s import *' % self.inputs['ConfigurationFile']"))
        self.log     = file(self.inputs['LogDirectory'] + '/pipeline_manager.log', 'a', 1)
        self.messages.addlogger(message.DebugLevel, self.log)
        self.print_message('----- Logging started -----')
        ExistingJobs = os.listdir(self.inputs['JobsDirectory'])
        self.server  = server
        self.client  = client
        for e in ExistingJobs:
            self.new_job(e)

    def communicate_job(self, job): 
        """function to write to log and communicate with GUI"""
        if   job['Status'] == JobError:     self.print_notification('Job:' + str(job['ExportID']) + ' Failed')
        elif job['Status'] == JobHold:      self.print_notification('Job:' + str(job['ExportID']) + ' is on Hold')
        elif job['Status'] == JobScheduled: self.print_notification('Job:' + str(job['ExportID']) + ' Scheduled')
        elif job['Status'] == JobProducing: self.print_notification('Job:' + str(job['ExportID']) + ' Started')
        elif job['Status'] == JobProduced:  self.print_notification('Job:' + str(job['ExportID']) + ' Produced')
        try:
            if not isinstance(self.client, types.NoneType):
                self.print_message(self.client.setStatus(str(job['ExportID']), str(job['Status'])))
        except: # this should be smarter to recover when the server comes back ?
            self.print_error('Could not update job %s status to %s.' % (str(job['ExportID']), str(job['Status'])))

    def new_job(self, filename):
        """Read filename and add to the list of jobs if it is a valid file."""
        import shutil
        try:
            shutil.move(self.inputs['NewJobsDirectory'] + '/' + filename, self.inputs['JobsDirectory'] + '/' + filename)
        except:
            self.print_debug('file not found (existing job?): ' + self.inputs['NewJobsDirectory'] + '/' + filename)
        self.parser.inputs['Job'] = self.inputs['JobsDirectory'] + '/' + filename
        self.parser.outputs       = ingredient.WSRTingredient() ## empty ingredient to be able to run more than once
        self.parser.go()
        job             = self.parser.outputs.copy()
        job['filename'] = filename
        if job['Status'] == JobScheduled:
            self.jobs.append(job)
            self.communicate_job(job)
        else:
            self.print_notification('Parsing ' + self.inputs['JobsDirectory'] + '/' + filename + ' failed') ## Not implemented yet

    def find_job(self, exportID):
        for j in self.jobs:
            if j['ExportID'] == exportID:
                return j
        return None

    def update_job(self, exportID, status): 
        """for communicating job status with GUI, mainly to put on Hold."""
        j = self.find_job(exportID)
        if j:
            j['Status'] = status
            self.communicate_job(j)
        else:
            self.print_debug('Job ' + str(exportID) + ' not found, ignoring message.')

    def check_server(self):
        """Check if there are any new jobs communicated to the server."""
        if not isinstance(self.server, types.NoneType):
            self.server.jobs_lock.acquire()
            while len(self.server.jobs) > 0:
                job = self.server.jobs.pop(0)
                if job[0] == NewJob:
                    self.new_job(job[1])
                elif job[0] == UpdateJob:
                    self.update_job(job[1], job[2])
            self.server.jobs_lock.release()

    def next_job(self):
        """See if there is another job scheduled."""
        import shutil, sys
        for j in self.jobs:
            if j['Status'] >= JobScheduled:
                self.idle = 0
                if j['Status'] > JobScheduled:
                    self.print_notification('Restating job: ' + j['ExportID'])
                j['Status'] = JobProducing
                self.communicate_job(j)
                self.cook_job(j)
                self.communicate_job(j)
                shutil.move(self.inputs['JobsDirectory'] + '/' + j['filename'], 
                            self.inputs['LogDirectory'] + '/' + str(j['ExportID']) + '/' + j['filename'])
                self.jobs.remove(j)
                return
        ## we only get here if there are no more jobs
        if self.idle:
            t = time.gmtime()
            sys.stdout.write('\r%04d-%02d-%02d %02d:%02d:%02d' % (t[0], t[1], t[2], t[3], t[4], t[5]))
            sys.stdout.flush()
        else:
            print 'pipeline_manager: No jobs available, waiting for next job...'
        self.idle = 1
        if not isinstance(self.server, types.NoneType):
            self.server.event.wait(60)
        else:
            time.sleep(60) # let's take a nap, not realy useful, as there is no server ?

    def prepare_recipe_parameters(self, job):
        """Prepare ingedients and message handler for the cook to cook the recipe."""
        import sys
        logfile = file(self.inputs['LogDirectory'] + '/' + str(job['ExportID']) + '/pipeline_manager.log', 'a', 1)
        messages = message.WSRTmessages()
        results  = ingredient.WSRTingredient()
        if self.messages.log[sys.stdout] == message.DebugLevel:
            messages.addlogger(message.DebugLevel, logfile)
            messages.setloglevel(message.DebugLevel, sys.stdout)
        else:
            messages.addlogger(message.VerboseLevel, logfile)
            messages.setloglevel(message.ErrorLevel, sys.stdout)
        inputs = job.copy()
        del inputs['script_name']
        del inputs['Status']
        del inputs['filename']
        return (logfile, inputs, results, messages)

    def cook_job(self, job):
        """This starts a recipe with the inputs as defined in the job file."""
        if not os.path.isdir(self.inputs['LogDirectory'] + '/' + str(job['ExportID'])):
            os.makedirs(self.inputs['LogDirectory'] + '/' + str(job['ExportID']))
        logfile, inputs, results, messages = self.prepare_recipe_parameters(job)
        try:
            c = self.cook_recipe(job['script_name'], inputs, results, messages)
        except Exception, e:
            messages.append(message.ErrorLevel, str(e))
            job['Status'] = JobError
            results       = None
        if results:
            job['Status'] += 1 # something more elaborate?
            messages.append(message.VerboseLevel, 'Results:')
            for o in results.keys():
                messages.append(message.VerboseLevel, str(o) + ' = ' + str(results[o]))
        else: # should a recipe always return something?
            messages.append(message.VerboseLevel, 'No Results!')
            job['Status'] = JobError
        logfile.close()

    def go(self):
        self.startup()
        try:
            while 1:
                try:
                    self.next_job()
                    self.check_server()
                except KeyboardInterrupt:
                    self.print_notification('Pipeline Manager: Keyboard interupt detected, asking user...')
                    reply = raw_input('Do you want to end the pipeline manager (y/n)?')
                    if 'y' in reply:
                        raise KeyboardInterrupt ('Pipeline Manager: User wants to end program')
                    else:
                        self.idle = 0
        except KeyboardInterrupt, k:
            self.print_notification(str(k))
        except Exception, inst:
            self.print_error('Pipeline Manager: Exception caught: ' + str(type(Exception)) + ' ' + str(inst))
            raise inst
        if not isinstance(self.server, types.NoneType): ## check if the server is alive
            self.server.stop_lock.acquire()
            self.server.stop = True ## tell the server to stop
            self.server.stop_lock.release()
            print '*** Please wait for the SOAP server to finish, this can take up to 60 seconds. ***'

## Stand alone execution code ------------------------------------------
if __name__ == '__main__':
    standalone = pipeline_manager()
    standalone.main()
